(* $Id: frontend.ml,v 1.13 2000/04/28 13:45:25 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

open Findlib;;

exception Usage;;

type mode =
     M_use | M_query | M_install | M_remove | M_compiler of string
;;


type psubst =
    Const of string
  | Percent of char
;;


let percent_subst spec s =
  (* spec = [ 'c', [ "ctext1"; "ctext2"; ... ];
   *          'd', [ "dtext1"; "dtext2"; ... ] ]
   * All occurrences of %c in the string s are replaced as specified in spec.
   * spec is an association list with the characters of the %-notation as keys
   * and lists of strings as values. The result is a list of strings containing
   * every combination of substituted values.
   *)

  let l = String.length s in
  let rec preprocess i j =
    if j<l then begin
      match s.[j] with
	'%' ->
	  if j+1<l then begin
	    let c = s.[j+1] in
	    Const(String.sub s i (j-i)) :: Percent c :: preprocess (j+2) (j+2)
	  end
	  else failwith "bad format string"
      |	_ ->
	  preprocess i (j+1)
    end
    else
      if i<j then
	[Const(String.sub s i (j-i))]
      else
	[]
  in

  let rec subst prefix l =
    match l with
      [] -> [prefix]
    | Const s :: l' ->
	subst (prefix ^ s) l'
    | Percent c :: l' ->
	let replacements =
	  try List.assoc c spec 
	  with Not_found -> failwith "bad format string" in
	List.flatten
	  (List.map
	     (fun replacement ->
	       subst (prefix ^ replacement) l')
	     replacements)
  in

  subst "" (preprocess 0 0)
;;
	       

let rec remove_dups l =
  match l with
    x :: l' ->
      if List.mem x l' then remove_dups l' else x::remove_dups l'
  | [] -> []
;;


let arg n =
  if n < Array.length Sys.argv then Sys.argv.(n) else raise Not_found
;;


let use_package prefix pkgnames =
  let pdirs =
    List.map
      (fun pname ->
        try
          "-I " ^ package_directory pname
        with
	  Not_found -> failwith ("Cannot find package " ^ pname))
      pkgnames
  in

  print_endline (prefix ^ String.concat " " pdirs)
;;



(************************* format expansion *************************)


let expand predicates eff_packages format =

    (* format:
     * %p         package name
     * %d         package directory
     * %v         version
     * %a         archive file(s)
     * %A         archive files as single string
     * %o         link option(s)
     * %O         link options as single string
     *)

  List.flatten
    (List.map
       (fun pkg ->
	 let dir =
	   try package_directory pkg 
	   with Not_found -> failwith ("package '" ^ pkg ^ "' not found")
	 in
	 let spec =
	   [ 'p',  [pkg];
             'd',  [dir];
	     'v',  [try package_property predicates pkg "version"
	            with Not_found -> "[unspecified]"];
	     'a',  Split.in_words 
	             (try package_property predicates pkg "archive"
		      with Not_found -> "");
	     'A',  [String.concat " "
		       (Split.in_words 
		          (try package_property predicates pkg "archive"
			   with Not_found -> ""))];
	     'o',  Split.in_words_ws
	             (try package_property predicates pkg "linkopts"
		      with Not_found -> "");
	     'O',  [String.concat " "
		       (Split.in_words_ws 
		          (try package_property predicates pkg "linkopts"
			   with Not_found -> ""))];
	   ] 
	 in
	 percent_subst spec format)
       eff_packages)
;;


(************************** QUERY SUBCOMMAND ***************************)

let query_package () =
  Arg.current := 1;

  let long_format =
    "package:    %p\nversion:    %v\narchive(s): %A\nlinkopts:   %O\nlocation:   %d\n" in
  let i_format =
    "-I %d" in
  let l_format =
    "-ccopt -L%d" in
  let a_format =
    "%a" in
  let o_format =
    "%o" in


  let predicates = ref [] in
  let format = ref "%d" in
  let separator = ref "\n" in
  let prefix = ref "" in
  let suffix = ref "\n" in
  let recursive = ref false in
  let descendants = ref false in

  let packages = ref [] in

  let append_predicate s =
    let pl = Split.in_words s in
    predicates := !predicates @ pl
  in


  Arg.parse
    [ "-predicates", Arg.String append_predicate,
                  "      specifies comma-separated list of assumed predicates";
      "-format", Arg.String (fun s -> format := s),
              "          specifies the output format";
      "-separator", Arg.String (fun s -> separator := s),
                 "       specifies the string that separates multiple answers";
      "-prefix", Arg.String (fun s -> prefix := s),
              "          a string printed before the first answer";
      "-suffix", Arg.String (fun s -> suffix := s),
              "          a string printed after the last answer";
      "-recursive", Arg.Set recursive,
                 "       select direct and indirect ancestors/descendants, too";
      "-descendants", Arg.Set descendants,
                   "     query descendants instead of ancestors (the default)";
      "-long-format", Arg.Unit (fun () -> format := long_format),
                   "     specifies long output format";
      "-i-format", Arg.Unit (fun () -> format := i_format),
                "        prints -I options for ocamlc";
      "-l-format", Arg.Unit (fun () -> format := l_format),
                "        prints -ccopt -L options for ocamlc";
      "-a-format", Arg.Unit (fun () -> format := a_format),
                "        prints names of archives to be linked in for ocamlc";
      "-o-format", Arg.Unit (fun () -> format := o_format),
                "        prints link options for ocamlc";
    ] 
    (fun p -> packages := !packages @ [p])
"usage: ocamlfind query [ -predicates <p>  | -format <f> |
                         -long-format     | -i-format   |
                         -l-format        | -a-format   |
			 -o-format        | -prefix <p> |
                         -separator <s>   | -suffix <s> |
                         -descendants     | -recursive  ] package ...";

    (* check packages: *)
    List.iter
      (fun pkg ->
	try
	  let _ = package_directory pkg in
	  ()
	with
	  Not_found ->
	    failwith ("package '" ^ pkg ^ "' not found"))
      !packages;


    let list_directory d =
      try
	let dd = Unix.opendir d in
	let rec read () =
	  try
	    let name = Unix.readdir dd in
	    if name <> "." & name <> ".." then
	      name :: read()
	    else
	      read()
	  with
	    End_of_file ->
	      Unix.closedir dd;
	      []
	in
	read()
      with
	Unix.Unix_error (code,_,_) ->
	  prerr_endline ("Warning: cannot read directory " ^ d ^ ": " ^
			 Unix.error_message code);
	  []
    in


    let eff_packages =
      if !recursive then begin
	if !descendants then begin
	  (* This is a hack as it violates the hierarchy of layers *)
	  let dirs = !Metacache.ocamlpath in
	  let subdirs = 
	    List.flatten
	      (List.map list_directory dirs) in
	  (* The 'subdirs' are possible package names *)
	  List.iter
	    (fun d ->
	      try
		(* Causes the package to be loaded into the cache: *)
		let _ = Metacache.requires !predicates d in ()
	      with
		Not_found -> ()
	      |	Failure s ->
		  prerr_endline ("Warning: " ^ s))
	    subdirs;
	  (* Get the descendants: *)
	  Metacache.users !packages
	end
	else
	  package_deep_ancestors !predicates !packages 
      end
      else
	!packages
    in

    let answers = expand !predicates eff_packages !format in

     print_string !prefix;
     print_string (String.concat !separator answers);
     print_string !suffix;
;;

	     
(**************** OCAMLC/OCAMLMKTOP/OCAMLOPT subcommands ****************)

type pass_file_t =
    Pass of string
  | Impl of string
  | Intf of string
;;


let ocamlc which () =
  Arg.current := 1;

  let switches = ref [] in
  let pass_options = ref [] in
  let pass_files = ref [] in

  let linkpkg = ref false in

  let packages = ref [] in
  let predicates = ref [] in
  let dontlink = ref [] in


  let add_switch name = 
    Arg.Unit (fun () -> 
                switches := name :: !switches;
                pass_options := !pass_options @ [name]) in
  let add_spec name =
    Arg.String (fun s -> pass_options := !pass_options @ [name; s]) in
  let add_pkg =
    Arg.String (fun s -> packages := !packages @ (Split.in_words s)) in
  let add_pred =
    Arg.String (fun s -> predicates := !predicates @ (Split.in_words s)) in
  let add_dontlink =
    Arg.String (fun s -> dontlink := !dontlink @ (Split.in_words s)) in

  Arg.parse
    (List.flatten
    [ [
      "-package", add_pkg,
               " <name>   Refer to package when compiling";
      "-linkpkg", Arg.Set linkpkg,
               "          Link the packages in";
      "-predicates", add_pred,
                  " <p>   Add predicate <p> when resolving package properties";
      "-dontlink", add_dontlink,
                " <name>  Do not link in package <name> and its ancestors";
      "-passopt", Arg.String (fun s -> pass_options := !pass_options @ [s]),
               " <opt>    Pass option <opt> directly to ocamlc/opt/mktop\nSTANDARD OPTIONS:";

      "-a", add_switch "-a",
         "                Build a library";
      "-c", add_switch "-c",
         "                Compile only (do not link)";
      "-cc", add_spec "-cc",
          " <comp>        Use <comp> as the C compiler and linker";
      "-cclib", add_spec "-cclib",
             " <opt>      Pass option <opt> to the C linker";
      "-ccopt", add_spec "-ccopt",
             " <opt>      Pass option <opt> to the C compiler and linker";
      ];
      if which = "ocamlopt" then  [
      "-compact", add_switch "-compact",
               "          Optimize code size rather than speed"
      ]	
      else [];
      if which <> "ocamlopt" then [
      "-custom", add_switch "-custom",
              "           Link in custom mode";
      "-g", add_switch "-g",
         "                Save debugging information";
      ] else [];
      [
      "-i", add_switch "-i",
         "                Print the types";
      "-I", add_spec "-I",
         " <dir>          Add <dir> to the list of include directories";
      "-impl", Arg.String (fun s -> pass_files := !pass_files @ [ Impl s ]),
            " <file>      Compile <file> as a .ml file";
      ]	;
      if which = "ocamlopt" then [
      "-inline", add_spec "-inline",
              " <n>       Set aggressiveness of inlining to <n>";
      ]	else [];
      [
      "-intf", Arg.String (fun s -> pass_files := !pass_files @ [ Intf s ]),
            " <file>      Compile <file> as a .mli file";
      "-intf-suffix", add_spec "-intf-suffix",
                   " <s>  Suffix for interface file (default: .mli)";
      "-intf_suffix", add_spec "-intf_suffix",
                   " <s>  same as -intf-suffix";
      "-labels", add_switch "-labels",
              "           Use commuting label mode";
      "-linkall", add_switch "-linkall",
               "          Link all modules, even unused ones";
      ]	;
      if which <> "ocamlopt" then [
      "-make-runtime", add_switch "-make-runtime",
                    "     Build a runtime system";
      "-make_runtime", add_switch "-make_runtime",
                    "     same as -make-runtime";
      ]	else [];
      [
      "-noautolink", add_switch "-noautolink",
                  "       Don't automatically link C libraries specif'd in .cma files";
      "-noassert", add_switch "-noassert",
                "         Do not compile assertion checks";
      "-o", add_spec "-o",
         " <file>         Set output file name to <file>";
      "-output-obj", add_switch "-output-obj",
                  "       Output a C object file instead of an executable";
      ];
      if which = "ocamlopt" then [
      "-p", add_switch "-p",
         "                Compile/link with profiling support for \"gprof\"
                       (implies -predicates gprof)";
      ]	else if which = "ocamlcp" then [
      "-p", add_spec "-p",
	 " [afilmt]       Profile constructs specified by argument:
      a  Everything
      f  Function calls
      i  if ... then ... else
      l  while, for
      m  match ... with
      t  try ... with";
      ]	else [];
      [
      "-pp", add_spec "-pp",
          " <command>     Pipe sources through preprocessor <command>";
      "-rectypes", add_switch "-rectypes",
                "         Allow arbitrary recursive types";
      ]	;
      if which = "ocamlopt" then [
      "-S", add_switch "-S",
         "                Keep intermediate assembly file";
      ]	 else [];
      [
      "-thread", add_switch "-thread",
              "           Use thread-safe standard library (implies -predicate mt)";
      "-unsafe", add_switch "-unsafe",
              "           No bounds checking on array and string access";
      ]	;
      if which <> "ocamlopt" then [
      "-use-runtime", add_spec "-use-runtime",
                   " <path>   Generate bytecode for the given runtime system";
      "-use_runtime", add_spec "-use_runtime",
                   "          same as -use-runtime";
      ]	 else [];
      [
      "-v", add_switch "-v",
         "                Print compiler version number";
      "-verbose", add_switch "-verbose",
               "          Print calls to external commands";
      "-w", add_spec "-w",
         " <flags>        Enable or disable warnings according to <flags>:
     A/a enable/disable all warnings
     C/c enable/disable suspicious comment
     F/f enable/disable partially applied function
     M/m enable/disable overriden methods
     P/p enable/disable partial match
     S/s enable/disable non-unit statement
     U/u enable/disable unused match case
     V/v enable/disable hidden instance variables
     X/x enable/disable all other warnings
     default setting is A (all warnings enabled)";
       "-", Arg.String (fun s -> pass_files := !pass_files @  [ Pass s ]),
	 " <file>          Treat <file> as a file name (even if it starts with `-')";
       ]	
    ])
    (fun s -> pass_files := !pass_files @ [ Pass s])
    ("usage: ocamlfind " ^ which ^ " [options] file ...");

  begin match which with
    "ocamlc"     -> predicates := "byte" :: !predicates;
  | "ocamlcp"    -> predicates := "byte" :: !predicates;
  | "ocamlmktop" -> predicates := "byte" :: "toploop" :: !predicates;
  | "ocamlopt"   -> predicates := "native" :: !predicates;
  | _            -> failwith "unsupported backend"
  end;

  if List.mem "-thread" !switches then begin
    predicates := "mt" :: !predicates;
    (* check type of threads: posix or bytecode *)
    try
      let type_of_threads = package_property [] "threads" "type_of_threads" in
      if type_of_threads = "posix" then
	predicates := "mt_posix" :: !predicates;
    with
      Not_found -> ()
  end;

  if List.mem "-p" !switches then
    predicates := "gprof" :: !predicates;

  if Findlib_config.ocaml_has_autolinking && 
     not (List.mem "-noautolink" !switches) 
  then
    predicates := "autolink" :: !predicates;

  let verbose = List.mem "-verbose" !switches in

  if verbose then 
    print_string ("Effective set of predicates: " ^ 
		  String.concat "," !predicates ^ "\n");

  (* check packages: *)
  List.iter
    (fun pkg ->
      try
	let _ = package_directory pkg in
	()
      with
	Not_found ->
	  failwith ("package '" ^ pkg ^ "' not found"))
    !packages;
  List.iter
    (fun pkg ->
      try
	let _ = package_directory pkg in
	()
      with
	Not_found ->
	  failwith ("package '" ^ pkg ^ "' not found"))
    !dontlink;


  let eff_packages =
    package_deep_ancestors !predicates !packages in

  let eff_dontlink =
    package_deep_ancestors !predicates !dontlink in

  let eff_link =
    List.flatten
      (List.map
	 (fun pkg -> if List.mem pkg !dontlink then [] else [pkg])
	 eff_packages) in


  let eff_packages_dl =
    remove_dups (List.map package_directory eff_packages) in

  let eff_link_dl =
    remove_dups (List.map package_directory eff_link) in

  let init_file_needed =
    List.mem "toploop" !predicates & List.mem "findlib" eff_link in

  let init_file_name =
    if init_file_needed then
      Filename.temp_file "findlib_init" ".ml"
    else
      ""
  in

  if init_file_needed then begin
    (* Generate initializer for "findlib_top.cma" *)
    let f = open_out_gen 
	      [Open_wronly; Open_trunc; Open_text]
	      0o777
	      init_file_name in
    try
      output_string f 
	("Topfind.don't_load [" ^ 
	 String.concat ";"
	   (List.map
	      (fun pkg -> "\"" ^ String.escaped pkg ^ "\"")
	      eff_link) ^
	 "];;\n");
      output_string f
	("Topfind.predicates := [" ^ 
	 String.concat ";"
	   (List.map
	      (fun pred -> "\"" ^ String.escaped pred ^ "\"")
	      !predicates) ^
	 "];;\n");
      close_out f
    with
      any ->
	close_out f;
	Sys.remove init_file_name;
	raise any
  end;

  if init_file_needed then 
    at_exit
      (fun () ->
	let tr f x = try f x with _ -> () in
	tr Sys.remove init_file_name;
	tr Sys.remove (Filename.chop_extension init_file_name ^ ".cmi");
	tr Sys.remove (Filename.chop_extension init_file_name ^ ".cmo");
      );

  let stdlibdir =
    (* normalized form of the stdlib directory *)
    let d = Findlib_config.ocaml_stdlib in
    if d <> "" & d.[String.length d - 1] = '/' then
      String.sub d 0 (String.length d - 1)
    else
      d
  in
  let stdlibdirslash = stdlibdir ^ "/" in

  let i_options =
    List.flatten
      (List.map
	 (fun pkgdir ->
	    if pkgdir = stdlibdir  or  pkgdir = stdlibdirslash then
	      []
	    else
	      [ "-I"; pkgdir;
		"-ccopt"; "-I" ^ pkgdir; ])
	 eff_packages_dl) in

  let l_options =
    List.flatten
      (List.map
	 (fun pkgdir ->
	    if pkgdir = stdlibdir  or  pkgdir = stdlibdirslash then
	      []
	    else
	      [ "-ccopt"; "-L" ^ pkgdir; ])
	 eff_link_dl) in

  let archives =
    List.flatten
      (List.map
	 (fun pkg -> 
	   let al = try package_property !predicates pkg "archive"
	            with Not_found -> "" in
	   List.map 
	     (fun a -> Filename.concat (package_directory pkg) a) 
	     (Split.in_words al)
	   @
	   (if init_file_needed & pkg = "findlib" then
	     [ init_file_name ] 
	    else
	      []))
	 eff_link) in

  let linkopts =
    List.flatten
      (List.map
	 (fun pkg -> 
	   let ol = try package_property !predicates pkg "linkopts"
	            with Not_found -> "" in
	   Split.in_words_ws ol)
	 (List.rev eff_link)) in


  let pass_files' =
    List.flatten
      (List.map
	 (function
	      Pass s ->
		if s.[0] = '-' 
		then [ "-"; String.sub s 1 (String.length s - 1) ]
		else [ s ]
	    | Impl s ->
		[ "-impl"; s ]
	    | Intf s ->
		[ "-intf"; s ]
	 )
	 !pass_files)
  in

  let arguments =
    !pass_options @
    i_options @
    (if !linkpkg then l_options else []) @
    (if !linkpkg then archives else []) @
    pass_files' @
    (if !linkpkg then linkopts else [])
  in

  if verbose then
    print_string ("+ " ^ which ^ " " ^ String.concat " " arguments ^ "\n");

  flush stdout;
  
  let pid =
    Unix.create_process 
      which 
      (Array.of_list (which :: arguments)) 
      Unix.stdin
      Unix.stdout
      Unix.stderr
  in

  let (_,status) = Unix.waitpid [] pid in
  begin
    match status with
      Unix.WEXITED 0 -> ()
    | Unix.WEXITED n ->
	if verbose then
	  print_string (which ^ " returned with exit code " ^ string_of_int n ^ "\n");
	exit n
    | Unix.WSIGNALED _ ->
	print_string (which ^ " got signal and exited\n");
	exit 2
    | Unix.WSTOPPED _ ->
	failwith "Your operating system does not work correctly"
  end;
;;
      

(************************************************************************)



let findlib() =
  try
    package_directory "findlib"
  with
    Not_found -> failwith "cannot find essential package 'findlib'"
;;


let mk_shell_command cmd parms =
  let rec sh_protect s k =
    if k < String.length s then
      let c = s.[k] in
       match c with
         '\''   -> "'\\''" ^ sh_protect s (k+1)
       | '\000' -> failwith "cannot pass null character"
       | _      -> String.make 1 c ^ sh_protect s (k+1)
    else
      ""
  in

  let rec form_command prefix parms =
    match parms with
      [] -> prefix
    | p :: parms' ->
        form_command (prefix ^ " '" ^ sh_protect p 0 ^ "'") parms'
  in

  form_command cmd parms
;;


let install_package pname files =
  let install_script = Filename.concat (findlib()) "install" in
  let sitelib = default_location() in
  let cmd = mk_shell_command install_script (sitelib::pname::files) in
  let rc = Sys.command cmd in
  if rc <> 0 then
    failwith "*** FAILURE WHILE INSTALLING PACKAGE ***"
;;


let remove_package pname =
  let remove_script = Filename.concat (findlib()) "remove" in
  let sitelib = default_location() in
  let cmd = mk_shell_command remove_script [sitelib; pname] in
  let rc = Sys.command cmd in
  if rc <> 0 then
    failwith "*** FAILURE WHILE REMOVING PACKAGE ***"
;;


let select_mode() =
  let m_string = try arg 1 with Not_found -> raise Usage in
  let m =
    match m_string with
      ("use"|"-use") ->          M_use
    | ("query"|"-query") ->      M_query
    | ("install"|"-install") ->  M_install
    | ("remove"|"-remove") ->    M_remove
    | ("ocamlc"|"-ocamlc") ->    M_compiler "ocamlc"
    | ("ocamlcp"|"-ocamlcp") ->  M_compiler "ocamlcp"
    | ("ocamlmktop"|"-ocamlmktop") -> M_compiler "ocamlmktop"
    | ("ocamlopt"|"-ocamlopt") -> M_compiler "ocamlopt"
    | _ -> raise Usage
  in

  m
;;


let main() =
  try
    let m = select_mode() in
    let l = Array.length Sys.argv in
    let rest = Array.sub Sys.argv 2 (l-2) in
    if rest = [| |] then raise Usage;
    match m with
      M_use      -> if rest.(0) = "-p" then begin
	              if l<4 then raise Usage;
	              use_package rest.(1) 
	                          (List.tl(List.tl(Array.to_list rest)))
	            end 
	            else
                      use_package "" (Array.to_list rest)
    | M_query    -> query_package ()
    | M_install  -> if l < 4 then raise Usage;
                    install_package (arg 2) (List.tl (Array.to_list rest))
    | M_remove   -> if l <> 3 then raise Usage;
                    remove_package (arg 2)
    | M_compiler which -> ocamlc which ()
  with
    Usage ->
      prerr_endline "usage: ocamlfind query [-help | other options] <package_name> ...";
      prerr_endline "   or: ocamlfind ocamlc [-help | other options] <file> ...";
      prerr_endline "   or: ocamlfind ocamlcp [-help | other options] <file> ...";
      prerr_endline "   or: ocamlfind ocamlmktop [-help | other options] <file> ...";
      prerr_endline "   or: ocamlfind ocamlopt [-help | other options] <file> ...";
      prerr_endline "   or: ocamlfind use [-p <prefix>] <package_name> ... (DEPRECATED)";
      prerr_endline "   or: ocamlfind install <package_name> <file> ...";
      prerr_endline "   or: ocamlfind remove <package_name>";
      exit 2
  | Failure f ->
      prerr_endline ("ocamlfind: " ^ f);
      exit 2
;;


try
  Sys.catch_break true;
  main()
with
  any ->
    prerr_endline ("Uncaught exception: " ^ Printexc.to_string any);
    exit 3
;;


(* ======================================================================
 * History:
 *
 * $Log: frontend.ml,v $
 * Revision 1.13  2000/04/28 13:45:25  gerd
 * 	The compiler frontends do not produce -I and -L options for the
 * standard library directory anymore.
 *
 * Revision 1.12  2000/04/26 00:09:20  gerd
 * 	O'Caml 3 changes.
 *
 * Revision 1.11  2000/02/28 20:23:53  gerd
 * 	Bugfix: option -output-obj works now.
 *
 * Revision 1.10  2000/02/28 20:22:05  gerd
 * 	Minor change.
 *
 * Revision 1.8  2000/01/10 22:48:04  gerd
 * 	The relative order of files passed using -intf <file> and
 * -impl <file> is not changed.
 *
 * Revision 1.7  1999/11/10 23:45:17  gerd
 * 	The -dontlink option can now list several packages.
 *
 * Revision 1.6  1999/07/09 15:28:38  gerd
 * 	Added automatic recognition of POSIX threads. The META file in
 * the "threads" package has now a variable "type_of_threads" with
 * possible values "bytecode" and "posix". This variable is set when
 * "findlib" is configured. The compiler frontends query this variable
 * (with an empty predicate list), and add "mt_posix" to the predicate
 * list if the variable has the value "posix".
 *
 * Revision 1.5  1999/06/26 15:44:33  gerd
 * 	Some minor changes; the /tmp files are now removed always;
 * multiple directories are only passed once to the underlying compiler.
 *
 * Revision 1.4  1999/06/26 15:01:49  gerd
 * 	Added the -descendants option.
 *
 * Revision 1.3  1999/06/24 20:17:50  gerd
 * 	Further modifications (dont know which...)
 *
 * Revision 1.2  1999/06/20 19:26:24  gerd
 * 	Major change: Added support for META files. In META files, knowlege
 * about compilation options, and dependencies on other packages can be stored.
 * The "ocamlfind query" subcommand has been extended in order to have a
 * direct interface for that. "ocamlfind ocamlc/ocamlopt/ocamlmktop/ocamlcp"
 * subcommands have been added to simplify the invocation of the compiler.
 *
 * Revision 1.1  1999/03/26 00:02:47  gerd
 * 	Initial release.
 *
 *
 *)
